#include <iostream>

#include <QApplication>
#include <QtCore>
#include <QDebug>
#include <QImage>
#include <QPrinter>
#include <QPixmap>
#include <QPainter>
#include <QRegExp>
#include <QtXml>

// Painter classes
class basepainter : public QPainter
{
public:
  basepainter (void) : 
    m_fname(""), m_scale(1.0)
  { }

  basepainter (QString fname, QRectF /*size*/, double scl) : 
    m_fname(fname), m_scale(scl)
  { }

  ~basepainter (void) { }

  QString get_fname (void) const { return m_fname; }

  double get_scale (void) const { return m_scale; }

  virtual void finish (void) = 0;

private:
  QString m_fname;
  double m_scale;
};

class pdfpainter : public basepainter
{
public:
  pdfpainter (QString fname, QRectF size, double scl) : 
    basepainter (fname, size, scl), m_printer()
  { 
    // Scale
    size.setHeight (scl * size.height ());
    size.setWidth (scl * size.width ());

    // Printer settings
    m_printer.setOutputFormat (QPrinter::PdfFormat);
    m_printer.setFontEmbeddingEnabled (true);
    m_printer.setOutputFileName (get_fname ());
    m_printer.setPageMargins (0.0F, 0.0F, 0.0F, 0.0F, QPrinter::DevicePixel);
    m_printer.setPaperSize (size.size (), QPrinter::DevicePixel);

    // Painter settings
    begin (&m_printer);
    resetTransform ();
    setViewport (size.toRect ());
    scale (scl, scl);
  }

  ~pdfpainter (void) { }

  void finish (void) { end (); }

private:
  QPrinter m_printer;
};

class rasterpainter : public basepainter
{
public:
  rasterpainter (QString fname, QRectF sizef, double scl, const char *fmt = 0): 
    basepainter (fname, sizef, scl), m_image (), m_fmt(fmt)
  {
    // Scale
    QRect size = sizef.toRect ();
    size.setHeight (scl * size.height ());
    size.setWidth (scl * size.width ());
    
    m_image = QPixmap (size.size ());

    // Painter settings
    begin (& m_image);
    resetTransform ();
    setViewport (size);
    scale (scl, scl);
  }

  ~rasterpainter (void) { }

  void finish (void) 
  {
    end (); 
    m_image.save (get_fname (), m_fmt);
  }

private:
  QPixmap m_image;
  const char *m_fmt;
};


// String conversion functions
QVector<float> qstr2vectorf (QString str)
{
  QVector<float> pts;
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 1)
    {
      float pt = (*p).toFloat ();
      pts.append (pt);
    }
  return pts;
}

QVector<double> qstr2vectord (QString str)
{
  QVector<double> pts;
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 1)
    {
      double pt = (*p).toDouble ();
      pts.append (pt);
    }
  return pts;
}

QVector<QPointF> qstr2ptsvector (QString str)
{
  QVector<QPointF> pts;
  str.replace (" ", ",");
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 2)
    {
      QPointF pt ((*p).toFloat (), (*(p+1)).toFloat ());
      pts.append (pt);
    }
  return pts;
}

QVector<QPoint> qstr2ptsvectord (QString str)
{
  QVector<QPoint> pts;
  str.replace (" ", ",");
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 2)
    {
      QPoint pt ((*p).toDouble (), (*(p+1)).toDouble ());
      pts.append (pt);
    }
  return pts;
}

// Extract field arguments in a style-like string, e.g. "bla field(1,34,56) bla"

QString get_field (QString str, QString field)
{
  QString retval;
  QRegExp rx (field + "\\(([^\\)]*)\\)");
  int pos = 0;
  pos = rx.indexIn (str, pos);
  if (pos > -1)
    retval = rx.cap (1);
  return retval;
}

// Polygon reconstruction class
class octave_polygon
{
public:
  octave_polygon (void) : 
    m_polygon(), m_color(QColor ("black")) 
  { }

  octave_polygon (QPolygonF p, QColor c) : 
    m_polygon(p), m_color(c) 
  { }

  ~octave_polygon (void) { }

  void draw (basepainter &painter, bool do_clear = false)
  {
    if (m_polygon.count () > 2)
      {
        painter.setBrush (m_color);
        painter.setPen (Qt::NoPen);

        painter.setRenderHint (QPainter::Antialiasing, false);
        painter.drawPolygon (m_polygon);
        painter.setRenderHint (QPainter::Antialiasing, true);
        if (do_clear)
          m_polygon.clear ();
      }
  }
  
  bool merge (QPolygonF p, QColor c)
  {
    if (m_polygon.count () == 0)
      {
        m_polygon = p;
        m_color = c;
        return true;
      }
      
    if (m_color != c)
      return false;

    bool retval = false;
    QPolygonF tmp = m_polygon.united (p);
    if (tmp.count () <= (m_polygon.count () + 30))
      {
        m_polygon = tmp;
        retval = true;
      }
    else
      qDebug () << "Counts: " << tmp.count ()
                << ".vs." << m_polygon.count () << "\n";

    return retval;
  }

private:
  QPolygonF m_polygon;
  QColor m_color;
};

void draw (QDomElement& elt, basepainter& painter)
{
  QDomNodeList nodes = elt.childNodes ();

  static QString clippath_id;
  static QMap< QString, QVector<QPoint> > clippath;

  static octave_polygon current_polygon;

  // tspan elements must have access to the font and position extracted from 
  // their paret text element
  static QFont font;
  static qreal dx = 0, dy = 0;

  for (int i = 0; i < nodes.count (); i++)
    {
      QDomNode node = nodes.at (i);
      if (! node.isElement ())
        return;

      QDomElement elt = node.toElement ();

      // Draw eventual remaining polygon
      if (elt.tagName () != "polygon")
        current_polygon.draw (painter, true);

      if (elt.tagName () == "clipPath")
        {
          clippath_id = "#" + elt.attribute ("id");
          draw (elt, painter);
          clippath_id = QString ();
        }
      else if (elt.tagName () == "g")
        {
          bool current_clipstate = painter.hasClipping ();
          QRegion current_clippath = painter.clipRegion ();

          QString str = elt.attribute ("clip-path");
          if (! str.isEmpty ())
            {
              QVector<QPoint> pts = clippath[get_field (str, "url")];
              if (! pts.isEmpty ())
                {
                  painter.setClipRegion (QRegion (QPolygon (pts)));
                  painter.setClipping (true);
                }
            }

          draw (elt, painter);

          // Draw eventual remaining polygon 
          current_polygon.draw (painter, true);

          // Restore previous clipping settings
          painter.setClipRegion (current_clippath);
          painter.setClipping (current_clipstate);
        }
      else if (elt.tagName () == "text")
        {
          // Font
          font = QFont ();
          font.setFamily (elt.attribute ("font-family"));

          if (elt.attribute ("font-weight") != "normal")
            font.setWeight (QFont::Bold);

          font.setPixelSize (elt.attribute ("font-size").toFloat ());
          painter.setFont (font);

          // Translation and rotation
          painter.save ();
          QString str = get_field (elt.attribute ("transform"), 
                                   "translate");
          if (! str.isEmpty ())
            {
              QStringList trans = str.split (",");
              dx = trans[0].toFloat ();
              dy = trans[1].toFloat ();
                  
              str = get_field (elt.attribute ("transform"), "rotate");
              if (! str.isEmpty ())
                {
                  QStringList rot = str.split (",");
                  painter.translate (dx+rot[1].toFloat (), 
                                     dy+rot[2].toFloat ());
                  painter.rotate (rot[0].toFloat ());
                  dx = rot[1].toFloat ();
                  dy = rot[2].toFloat ();
                }
              else
                {
                  painter.translate (dx, dy);
                  dx = 0;
                  dy = 0;
                }
            }
              
          draw (elt, painter);
          painter.restore ();
        }
      else if (elt.tagName () == "tspan")
        {
          // Color is specified in rgb
          QString str = get_field (elt.attribute ("fill"), 
                                   "rgb");
          if (! str.isEmpty ())
            {
              QStringList clist = str.split (",");
              painter.setPen (QColor (clist[0].toInt (), clist[1].toInt (),
                                      clist[2].toInt ()));
            }

          QStringList xx = elt.attribute ("x").split (" ");
          int x = xx[0].toInt ();
          int y = elt.attribute ("y").toInt ();
          painter.drawText (x-dx, y-dy, elt.text ());
          draw (elt, painter);
        }
      else if (elt.tagName () == "polyline")
        {
          // Color
          QColor c (elt.attribute ("stroke"));
          QString str = elt.attribute ("stroke-opacity");
          if (! str.isEmpty () && str.toFloat () != 1.0F 
              && str.toFloat () >= 0.0F)
            c.setAlphaF (str.toFloat ());
          
          QPen pen;
          pen.setColor (c);

          // Line properies
          float w = elt.attribute ("stroke-width").toFloat ();
          pen.setWidth (w);

          str = elt.attribute ("stroke-linecap");
          pen.setCapStyle (Qt::SquareCap);
          if (str == "round")
            pen.setCapStyle (Qt::RoundCap);
          else if (str == "butt")
            pen.setCapStyle (Qt::FlatCap);
              
          str = elt.attribute ("stroke-linejoin");
          pen.setJoinStyle (Qt::MiterJoin);
          if (str == "round")
            pen.setJoinStyle (Qt::RoundJoin);
          else if (str == "bevel")
            pen.setJoinStyle (Qt::BevelJoin);

          str = elt.attribute ("stroke-dasharray");
          pen.setStyle (Qt::SolidLine);
          if (! str.isEmpty ())
            {
              QVector<double> pat = qstr2vectord (str);
              if (pat.count () != 2 || pat[1] != 0)
                pen.setDashPattern (qstr2vectord (str));
            }

          painter.setPen (pen);
          painter.drawPolyline (qstr2ptsvector (elt.attribute ("points")));
        }
      else if (elt.tagName () == "image")
        {
          // Images are represented as a base64 stream of png formated data
          QString href_att = elt.attribute ("xlink:href");
          QString prefix ("data:image/png;base64,");
          QByteArray data = 
            QByteArray::fromBase64(href_att.mid (prefix.length ())
                                   .toLocal8Bit ());
          QImage img;
          if (img.loadFromData (data, "PNG"))
            {
              QRect pos(elt.attribute ("x").toInt (),
                        elt.attribute ("y").toInt (),
                        elt.attribute ("width").toInt (),
                        elt.attribute ("height").toInt ());

              // Translate
              painter.save ();
              QString str = get_field (elt.attribute ("transform"), "matrix");
              if (! str.isEmpty ())
                {
                  QVector<float> m = qstr2vectorf (str);
                  double scl = painter.get_scale ();
                  QTransform tform(m[0]*scl, m[1]*scl, m[2]*scl, 
                                   m[3]*scl, m[4]*scl, m[5]*scl);
                  painter.setTransform (tform);
                }
                  
              painter.drawImage (pos, img);
              painter.restore  ();
            }               
        }
      else if (elt.tagName () == "polygon")
        {
          if (! clippath_id.isEmpty ())
            clippath[clippath_id] = qstr2ptsvectord (elt.attribute ("points"));
          else
            {
              QString cs = elt.attribute ("fill");
              if (! cs.isEmpty ())
                {
                  QColor c = QColor (cs);

                  QString str = elt.attribute ("fill-opacity");
                  if (! str.isEmpty () && str.toFloat () != 1.0F 
                      && str.toFloat () >= 0.0F)
                    c.setAlphaF (str.toFloat ());

                  QPolygonF p (qstr2ptsvector (elt.attribute ("points")));
                  if (! current_polygon.merge (p, c))
                    {
                      // Draw previous polygon
                      current_polygon.draw (painter);
                      current_polygon = octave_polygon (p, c);
                    }
                }
            }
        }
    }
}

int main(int argc, char *argv[])
{  
  const char *help = 
"Usage:\n\
\tqt-svg-convert infile fmt dpi outfile\n\n\
Convert svg file to pdf, png or jpg file. All argument are mandatory:\n\
* infile: input svg file\n\
* dpi: device dependent resolution in screen pixel per inch\n\
* fmt: format of the output file. May be one of pdf, bmp, jpg, png or ppm\n\
* outfile: output file name\n";

  if (argc != 5)
    {
      printf ("%s", help);
      return -1;
    }

  // Create a DOM document 
  QDomDocument document;

  // Open a file for reading
  QFile file(argv[1]);
  if(!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
      std::cout << "Unable to open file " << argv[1] << std::endl;
      printf ("%s", help);
      return -1;
    }
  else
    {
      // Load xml file
      if(!document.setContent(&file))
        {
          std::cout << "Failed to load the file for reading.";
          printf ("%s", help);
          return -1;
        }
      file.close();
    }

  // Format and resolution
  if (strcmp (argv[2], "pdf") && strcmp (argv[2], "bmp") 
      && strcmp (argv[2], "jpg") && strcmp (argv[2], "png")
      && strcmp (argv[2], "ppm"))
    {
      printf ("Unrecognized  output file format %s\n%s", argv[2], help);
      return -1;
    }

  float dpi = QString (argv[3]).toFloat ();
  if (dpi <= 0.0F)
    {
      printf ("dpi must be positive");
      return -1;
    }


  QApplication a(argc, argv);

  // Get viewport from the root element
  QDomElement root = document.firstChildElement();
  int x0, y0, dx, dy;
  QString s = root.attribute ("viewBox");
  QTextStream (&s) >> x0 >> y0 >> dx >> dy;

  QRectF vp (x0, y0, dx, dy);
  float scl = dpi / 72.0F;

  if (! strcmp (argv[2], "pdf"))
    {
      // PDF painter
      pdfpainter painter (argv[4], vp, scl);
      draw (root, painter);
      painter.finish ();
    }
  else
    {
      // Raster painter
      rasterpainter rpainter (argv[4], vp, scl, argv[2]);
      draw (root, rpainter);
      rpainter.finish ();
    }
  
  return 0;
}
